@@ -2,7 +2,7 @@
 #
 # MANAGESIEVE (timsieved) client script
 #
-# Copyright © 2006-2013 Phil Pennock.  All rights reserved.
+# Copyright © 2006-2016 Phil Pennock.  All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without
 # modification, are permitted provided that the following conditions
@@ -52,6 +52,19 @@ my %ssl_options = (
 # These defaults can be overriden on the cmdline:
 my ($forbid_clearauth, $forbid_clearchan) = (0, 0);
 
+# Note Well: if SSL/TLS verification is enabled, that is equivalent to
+# $forbid_clearchan, since no TLS means no verification possible.  But
+# if verification is not required then by default clearchan is allowed, not
+# just unverified TLS.  The command-line overrides these.
+# This is a breaking change between v0.87 and v0.88 as these semantics were
+# firmed up and --clearchan was added to restore the old default behaviour
+# if TLS was never offered but verification was enabled.
+#
+# I expect this to cause complaints as it's not backwards compatible, but the
+# old behaviour was susceptible to MitM connection downgrade by removing the
+# capability advertisement.  The new behaviour is a security improvement, and
+# the old semantics can be obtained by explicitly setting --clearchan.
+
 # Add a key to this to blacklist that authentication mechanism.  Might be
 # useful on some platforms with broken libraries.  Make sure the key is
 # upper-case!
@@ -77,7 +90,7 @@ my $OPENSSL_COMMAND = 'openssl';
 # ######################################################################
 # No user-serviceable parts below
 
-our $VERSION = '0.87';
+our $VERSION = '0.88';
 
 use Authen::SASL 2.11 qw(Perl);
 # 2.11: first version with non-broken DIGEST-MD5
@@ -172,6 +185,8 @@ sub closedie_NOmsg;
 sub die_NOmsg;
 sub fixup_ssl_configuration;
 sub derive_sieve_server;
+sub escalate_warnings_to_exitcode;
+sub exit_with_result_based_on_escalated_warnings;
 
 my $DEBUGGING_SASL = 0;
 my $DATASTART = tell DATA;
@@ -180,6 +195,9 @@ my $remotesievename;
 my $port = undef;
 my ($user, $authzid, $authmech, $sslkeyfile, $sslcertfile, $ssl_cert_fingerprint, $passwordfd);
 my ($tlscapath, $tlscafile);
+my $tls_explicit_hostname = undef;
+my $tls_sufficiently_configured = 0;
+my $clearchan_explicitly_set = 0;
 my $prioritise_auth_external = 0;
 my $dump_tls_information = 0;
 my $opt_version_req = 0;
@@ -189,6 +207,7 @@ my ($server, $realm);
 my $net_domain = AF_UNSPEC;
 my $action = 'command-loop';
 my $execscript;
+
 GetOptions(
 	# settings which adjust how we connect
 	"localsieve=s"	=> \$localsievename,
@@ -208,8 +227,10 @@ GetOptions(
 	"tlscertfingerprint|sslcertfingerprint=s" => \$ssl_cert_fingerprint,
 	"tlscapath=s"	=> \$tlscapath,
 	"tlscafile=s"	=> \$tlscafile,
+	"tlshostname=s"	=> \$tls_explicit_hostname,
 	"noclearauth"	=> \$forbid_clearauth,
-	"noclearchan"	=> sub { $forbid_clearauth = $forbid_clearchan = 1 },
+	"noclearchan"	=> sub { $clearchan_explicitly_set = $forbid_clearauth = $forbid_clearchan = 1 },
+	"clearchan"	=> sub { $forbid_clearchan = 0; $clearchan_explicitly_set = 1; },
 	"4"		=> sub { $net_domain = AF_INET },
 	"6"		=> sub { $net_domain = AF_INET6 },
 	"debug"		=> \$DEBUGGING,
@@ -316,7 +337,7 @@ if (defined $sslkeyfile) {
 }
 
 if (defined $localsievename and not defined $remotesievename) {
-	$remotesievename = $localsievename;
+	$remotesievename = File::Basename::basename($localsievename);
 }
 
 if (defined $localsievename and $action =~ /upload|checkscript/) {
@@ -336,6 +357,25 @@ if ($action eq 'deactivate' and defined $remotesievename) {
 	# conceivably useful, so ignored at least for the present.
 }
 
+# This happens after fixup_ssl_configuration() above, so we've gone to
+# some lengths to get CA anchors available already.
+if (exists $ssl_options{SSL_ca_path} or exists $ssl_options{SSL_ca_file}) {
+	$tls_sufficiently_configured = 1;
+}
+if ($ssl_options{'SSL_verify_mode'} != 0x00) {
+	unless (exists $ENV{SIEVECONNECT_INSECURE_CLEARTEXT_FALLBACK} and
+		length $ENV{SIEVECONNECT_INSECURE_CLEARTEXT_FALLBACK}) {
+		$forbid_clearchan = 1 unless $clearchan_explicitly_set;
+	}
+
+	if ($forbid_clearchan and not $tls_sufficiently_configured) {
+		warn("TLS verification is ON, so non-TLS is DISABLED\n");
+		warn(" but we're MISSING configuration information for CA anchors.\n");
+		warn(" No secure connection possible, aborting early.\n");
+		die("fix TLS setup, else use --noclearchan or --notlsverify\n");
+	}
+}
+
 # ######################################################################
 # Start work; connect, start TLS, authenticate
 
@@ -343,6 +383,7 @@ if ($action eq 'deactivate' and defined $remotesievename) {
 # for us in future. So, TLS negotiations do not use insecure hostnames from
 # DNS.
 my $trusted_server = $server;
+$trusted_server = $tls_explicit_hostname if defined $tls_explicit_hostname;
 
 my @host_port_pairs;
 
@@ -590,7 +631,13 @@ parse_capabilities $sock;
 
 my $tls_bitlength = -1;
 
-if (exists $capa{STARTTLS}) {
+# This will be called immediately after function definition if
+# exists $capa{STARTTLS}; moved into a function to make it easy to abort when TLS
+# is not mandatory.
+#
+# Return 1 if all okay, 0 if TLS not configured, or die on issues.
+sub handle_capa_STARTTLS
+{
 	# Always set SSL_hostname for SNI, only set the verification
 	# modes if not disabled.
 	$ssl_options{'SSL_hostname'} = $trusted_server;
@@ -599,6 +646,22 @@ if (exists $capa{STARTTLS}) {
 		# we want full wildcard support in same style that most admins
 		# are already familiar with:
 		$ssl_options{'SSL_verifycn_scheme'} = 'http';
+
+		unless ($tls_sufficiently_configured) {
+			debug("-T- offered STARTTLS, want verification, TLS insufficiently configured to proceed");
+			warn("offered STARTTLS, no CA anchor information available\n");
+			if ($forbid_clearchan) {
+				# we shouldn't even have tried to establish a connection in this scenario,
+				# but protect against screw-ups
+				debug("-T- we should not even be here (BUG please report!)");
+				warn("cleartext communications disabled, no TLS verification possible\n");
+				die("fix TLS setup, else use --clearchan or --notlsverify\n");
+			}
+			# to be at this line of code, in theory the invoker must have explictly set --clearchan
+			# (--notlsverify would skip this entire block).
+			return 0;
+		}
+
 	}
 
 	if ($DEBUGGING) {
@@ -696,12 +759,34 @@ if (exists $capa{STARTTLS}) {
 	# This means that if they don't support NOOP by 2.3.14, I have to
 	# figure out how to decide what is safe and backtrack which version
 	# precisely was the first to send the capability response correctly.
+	#
+	# There is also at least one special version which needs special
+	# treatment despite having a higher version number: Kolab's patched
+	# "nocaps" Cyrus version (see below).
 	my $use_noop = 1;
-	if (exists $capa{"IMPLEMENTATION"} and
-		$capa{"IMPLEMENTATION"} =~ /^Cyrus timsieved v2\.3\.(\d+)\z/ and
-		$1 >= 13) {
+	my $timsieved_nocaps = 0;
+	my $cyrus_version_extension = '';
+	if (exists $capa{"IMPLEMENTATION"}) {
+	    if ($capa{"IMPLEMENTATION"} =~ /^Cyrus timsieved v2\.3\.(\d+)\z/
+		    and $1 >= 13) {
 		debug("--- Cyrus drops connection with dubious log msg if send NOOP, skip that");
 		$use_noop = 0;
+	    } elsif ($capa{"IMPLEMENTATION"} =~
+		    /^Cyrus timsieved v2\.3\.\d+-([\w-]*nocaps)\z/ ) {
+		$cyrus_version_extension = $1;
+		# Special case: Cyrus may have been patched to not resend its
+		# capabilities for compatibility with older Kontact versions.
+		# Kolab does this, for example; see
+		# https://roundup.kolab.org/issue2443.  This patch may even
+		# have been applied to more modern versions such as v2.3.16.
+		# For this reason, we need to specifically check for this case.
+		# If a nocaps server is detected, nothing needs to be done (see
+		# the $timsieved_nocaps check below).
+		debug("--- Special Cyrus server version detected:" .
+			" $cyrus_version_extension");
+		$use_noop = 0;
+		$timsieved_nocaps = 1;
+	    }
 	}
 
 	if ($use_noop) {
@@ -710,6 +795,10 @@ if (exists $capa{STARTTLS}) {
 		parse_capabilities($sock,
 			sent_a_noop	=> $noop_tag,
 			external_first	=> $prioritise_auth_external);
+	} elsif ($timsieved_nocaps) {
+	    # For the nocaps version of Cyrus, nothing should be done here.
+	    debug("--- $cyrus_version_extension: using capabilities" .
+		    " transmitted before STARTTLS!");
 	} else {
 		parse_capabilities($sock,
 			external_first	=> $prioritise_auth_external);
@@ -719,8 +808,16 @@ if (exists $capa{STARTTLS}) {
 		parse_capabilities($sock,
 			external_first => $prioritise_auth_external);
 	}
-} elsif ($forbid_clearchan) {
-	die "TLS not offered, SASL confidentiality not supported in client.\n";
+	return 1;
+}
+
+my $tls_was_setup_okay = 0;
+if (exists $capa{STARTTLS}) {
+	$tls_was_setup_okay = handle_capa_STARTTLS();
+}
+
+if ($forbid_clearchan and not $tls_was_setup_okay) {
+	die "TLS not established, SASL confidentiality not supported in client.\n";
 }
 
 my %authen_sasl_params;
@@ -1036,6 +1133,14 @@ my %sieve_commands = (
 		params => 0,
 	},
 );
+if ($DEBUGGING) {
+	$sieve_commands{'debug'} = {
+		routine => \&aux_toggle_debugging,
+		help => 'debugging on or off',
+		params => 0,
+		params_max => 1,
+	};
+}
 
 my %subst_patterns = (
 	DATE		=> sub { return strftime '%Y-%m-%d', gmtime() },
@@ -1109,6 +1214,7 @@ unless ($ignore_server_version) {
 # apply, since GetOptions() sets $action for us.
 # 
 if ($action ne 'command-loop' and exists $sieve_commands{$action}{action}) {
+	escalate_warnings_to_exitcode();
 	closedie $sock, "internal error, no routine for \'$action\'"
 		unless exists $sieve_commands{$action}{routine};
 	my @params;
@@ -1131,7 +1237,7 @@ if ($action ne 'command-loop' and exists $sieve_commands{$action}{action}) {
 		$saveddie =~ s/^QUIT:\n?//;
 		die $saveddie if length $saveddie;
 	}
-	exit 0;
+	exit_with_result_based_on_escalated_warnings();
 }
 
 if ($action ne 'command-loop') {
@@ -1148,6 +1254,7 @@ $SIG{'PIPE'} = sub {
 };
 
 if (defined $execscript) {
+	escalate_warnings_to_exitcode();
 	$report_lineno = 1;
 	my $scripth = new IO::File $execscript, '<'
 		or closedie $sock, "Unable to read-open($execscript): $!\n";
@@ -1311,7 +1418,16 @@ $cmdlinedone_func->() if defined $cmdlinedone_func;
 
 sfinish $sock;
 print "\n";
-exit $exitval;
+# We can exit non-zero two ways:
+#  1. our tracked exitval here
+#  2. if nothing was serious enough to be an error for interactive use but
+#     triggered warnings, then we should have set up a warn-handler to bump
+#     a count, so we can emit a message with a count, then exit 1.
+#     * nb: it's safe to call this even if the non-interactive setup routine
+#       is never called, because the value will just stay stuck at 0 and we
+#       degenerate to `exit 0;`
+exit $exitval if $exitval;
+exit_with_result_based_on_escalated_warnings();
 
 # ######################################################################
 # The sieve commands.
@@ -1611,6 +1727,24 @@ sub aux_list_keywords
 	print "\t\%$_\n" foreach sort keys %subst_patterns;
 }
 
+sub aux_toggle_debugging
+{
+	my ($sock, $toggledir) = @_;
+	unless (defined $toggledir and length $toggledir) {
+		$DEBUGGING = $DEBUGGING ? 0 : 1;
+		print "Debugging set to: $DEBUGGING\n";
+		return;
+	}
+	if (grep {$_ eq lc($toggledir)} qw(yes on true enable 1)) {
+		$DEBUGGING = 1;
+	} elsif (grep {$_ eq lc($toggledir)} qw(no off false disable 0)) {
+		$DEBUGGING = 0;
+	} else {
+		print "Unrecognised debugging value: $toggledir\n";
+	}
+	print "Debugging set to: $DEBUGGING\n";
+}
+
 # ######################################################################
 # Term::ReadLine support.
 
@@ -1692,6 +1826,30 @@ sub complete_rl_sieve
 	}
 }
 
+# ######################################################################
+# Warnings escalation if not interactive
+
+{
+	my $seen_warnings = 0;
+
+	sub escalate_warnings_to_exitcode
+	{
+		$SIG{__WARN__} = sub {
+			$seen_warnings++;
+			# perldoc -f warn explicitly states that it's safe to call warn
+			# from within a __WARN__ handler.
+			warn(@_);
+		};
+	}
+
+	sub exit_with_result_based_on_escalated_warnings
+	{
+		exit 0 unless $seen_warnings;
+		warn("saw $seen_warnings problems, exiting 1\n");
+		exit 1;
+	}
+}
+
 # ######################################################################
 # minor routines
 
@@ -1987,6 +2145,7 @@ sub fixup_ssl_configuration
 		debug("setup: Have set SSL_ca_path to $ssl_options{'SSL_ca_path'}");
 		return;
 	}
+	debug("setup: FAILED to find SSL CA configuration, this will end badly");
 }
 
 # returns 1 "okay", 0 "definitely not available", undef "unknown"
@@ -2054,7 +2213,7 @@ sub derive_sieve_server
 		return (undef, undef, undef) unless $have_mozilla_public_suffix;
 	}
 
-	if ($have_mozilla_public_suffix) {
+	if ($have_mozilla_public_suffix and defined public_suffix($host)) {
 		$domain = $host;
 		debug "findserver: walking up hostname domains with Mozilla PSL";
 		while ($domain ne public_suffix($domain)) {
@@ -2074,7 +2233,7 @@ sub derive_sieve_server
 				last;
 			}
 		}
-	} else {
+	} elsif (defined $domain) {
 		debug "findserver: checking SRV records for: $domain";
 		my $have_srv = domain_sieve_server $domain;
 		if (defined $have_srv) {
@@ -2085,6 +2244,9 @@ sub derive_sieve_server
 			debug "findserver: SRV record explicitly says 'no service'";
 			return (undef, undef, 1); # force off SRV lookups
 		}
+	} else {
+		# $domain not set, so not "host.domain.tld", so probably host.local
+		debug "findserver: no public suffix and not in a host within a public domain";
 	}
 
 	# We give up, say nothing
@@ -2140,7 +2302,8 @@ sieve-connect - managesieve command-line client
                [--notlsverify|--nosslverify]
                [--tlscertfingerprint|--sslcertfingerprint <dgsttype:digest>]
                [--tlscapath <ca_directory>]|[--tlscafile <ca_file>]
-               [--noclearauth] [--noclearchan]
+               [--tlshostname <hostname>]
+               [--noclearauth] [--noclearchan] [--clearchan]
                [--authmech <mechanism>]
                [--ignoreserverversion]
                [--upload|--download|--list|--delete|--checkscript|--edit|
@@ -2201,6 +2364,11 @@ If that option is not present, then C<$IMAP_SERVER> from the environment is
 checked and, if it's not a unix-domain socket path, is used with any port
 specification stripped off.
 
+For TLS verification, this is the default name used for hostnames (both SNI and
+verification); no information derived from DNS is currently used as the trusted
+hostname identifier.  (This is subject to change in future, given DNSSEC).  The
+B<--tlshostname> option can be used to override the name used for TLS.
+
 Next, unless B<--nosrv> is given, checks are made for SRV records so as to search
 for a default server; if the F<Mozilla::PublicSuffix> Perl module is available,
 these checks are done for every level of the hostname upto (but not including)
@@ -2261,6 +2429,11 @@ active attacks and you are unable to arrange for the relevant Certificate
 Authority certificate to be available, then you can lower your safety with the
 B<--notlsverify> option, also spelt B<--nosslverify>.
 
+If verification is requested (the default) but TLS is not available, we
+do not fall back to cleartext insecure communications.  Use B<--clearchan>
+to change that, or set C<$SIEVECONNECT_INSECURE_CLEARTEXT_FALLBACK> non-empty
+in the environment.
+
 If you don't want to (only) rely on CA systems you can explicitly set an
 expected server certificate fingerprint using the B<--tlscertfingerprint>
 option, also spelt B<--sslcertfingerprint>.  If you wish to ignore CA
@@ -2332,6 +2505,36 @@ a "VERSION" capability, see B<--ignoreserverversion> to override.
 (If B<--server> is not explicitly stated, it may be provided at the end of
 the command-line for compatibility with sieveshell.)
 
+=head1 EXAMPLES
+
+Connect to a Sieve server and enter interactive mode, when you already have a
+Kerberos ticket and GSSAPI/Kerberos is available:
+
+    $ sieve-connect --server imap.example.org
+    ReadLine support enabled.
+    >
+
+Do the same, but with C<$IMAP_SERVER> set in environ:
+
+    $ sieve-connect
+    ReadLine support enabled.
+    >
+
+Upload a script from the current directory, being prompted to authenticate;
+note that the script won't be activated (uploading just makes it available,
+possibly with the server having first checked it for errors):
+
+    $ sieve-connect --server imap.example.org --user fred@example.org \
+         --localsieve fred.siv --upload
+    Sieve/IMAP Password: [password here, not shown]
+    $
+
+See a lot of what's happening under the covers:
+
+    $ sieve-connect --debug
+    [ snip 30 or so lines ]
+    >
+
 =head1 ENVIRONMENT
 
 C<$IMAP_SERVER> for a default IMAP server.
@@ -2341,6 +2544,9 @@ C<$USERNAME> and C<$LOGNAME> where the C<getpwuid()> function is not available.
 C<$SSL_CERT_DIR> and C<$SSL_CERT_FILE> for locating default
 Certificate Authority trust anchors.
 
+C<$SIEVECONNECT_INSECURE_CLEARTEXT_FALLBACK> to preserve old poor hygiene
+around TLS fallback.
+
 C<$VISUAL>, else C<$EDITOR>, for the edit action.
 
 =head1 BUGS
@@ -2403,6 +2609,16 @@ Versions prior to 0.85 did not actually verify the peer certificate identity,
 although this author stupidly believed that it did.
 API/expectations mismatch.
 
+Versions prior to 0.88 defaulted to falling back to cleartext in the absence
+of STARTTLS if CA information was configured locally and verification
+requested (the default).
+Today, this is no longer acceptable for client-server communications; either
+verify-and-require-TLS or don't-verify-and-fallback-to-cleartext.
+This is the new policy going forward; use B<--clearchan> to allow fallback
+while still trying to verify TLS (but why?) or B<--notlsverify> to skip
+verification.  Or add C<$SIEVECONNECT_INSECURE_CLEARTEXT_FALLBACK> non-empty
+in the environment to avoid the implicit noclearchan-when-verify-enabled.
+
 =head1 AUTHOR
 
 Phil Pennock E<lt>phil-perl@spodhuis.orgE<gt> is guilty, m'Lud.
@@ -2412,6 +2628,22 @@ at L<http://mail.globnix.net/mailman/listinfo/sieve-connect-announce> or you
 can send mail,
 L<mailto:sieve-connect-announce-request@spodhuis.org?subject=subscribe>
 
+=head1 AVAILABILITY
+
+Releases are made available at
+L<http://people.spodhuis.org/phil.pennock/software/> in the form of a tarball
+and an associated detached PGP signature.  All releases are signed, always, and
+always have been.  The signing key is in the PGP Strong Set (which means
+there's a stronger chance that you can verify the identity of the key owner).
+Historically, releases were signed with key C<0x403043153903637F>.  If you're
+reading this text from a release, then I've cut a new release since switching
+to key C<0x4D1E900E14C1CC04> and I expect that 4096RSA key to be used, barring
+major incident.
+
+The source code is available via Git; the authoritative public-facing
+repository is currently L<https://github.com/philpennock/sieve-connect> and
+pull-requests and bug-reports are accepted there.
+
 =head1 PREREQUISITES
 
 Perl.  F<Authen::SASL>.  F<IO::Socket::INET6>.
